# frozen_string_literal: true

# Redmine - project management software
# Copyright (C) 2006-2019  Jean-Philippe Lang
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

class ContractQuery < Query
  self.queried_class = Contract
  self.view_permission = :view_contracts

  self.available_columns = [
    QueryColumn.new(:contract_identifier, :sortable => "#{Contract.table_name}.contract_identifier", :groupable => true),
    QueryColumn.new(:contract_type, :sortable => "#{Contract.table_name}.contract_type", :groupable => true),
    QueryColumn.new(:project, :sortable => "#{Project.table_name}.name", :groupable => true),
    QueryColumn.new(:user, :sortable => lambda { User.fields_for_order_statement }, :groupable => true),
    QueryColumn.new(:counter_party, :sortable => lambda { Entity.fields_for_order_statement }, :groupable => true),
    QueryColumn.new(:date_valid_from, :sortable => "#{Contract.table_name}.date_valid_from", :groupable => true),
    QueryColumn.new(:date_valid_to, :sortable => "#{Contract.table_name}.date_valid_to", :groupable => true),
    QueryColumn.new(:reward_unit, :sortable => "#{Contract.table_name}.reward_unit", :groupable => true),
    QueryColumn.new(:unit_price, :sortable => "#{Contract.table_name}.unit_price", :groupable => true),
    QueryColumn.new(:original_currency, :sortable => "#{Contract.table_name}.original_currency", :groupable => true),
    QueryColumn.new(:original_unit_price, :sortable => "#{Contract.table_name}.original_unit_price", :groupable => true),
    QueryColumn.new(:exchange_rate, :sortable => "#{Contract.table_name}.exchange_rate", :groupable => true),
  ]

  def initialize(attributes = nil, *args)
    super attributes
    self.filters ||= { "contract_identifier" => { :operator => "*", :values => [] } }
  end

  def initialize_available_filters
    add_available_filter(
      "contract_identifier", :type => :text,
    )
    add_available_filter(
      "contract_type",
      :type => :list, :values => lambda { contract_type_values },
    )
    add_available_filter(
      "project_id",
      :type => :list, :values => lambda { project_values },
    ) if project.nil?
    if project && !project.leaf?
      add_available_filter(
        "subproject_id",
        :type => :list_subprojects,
        :values => lambda { subproject_values },
      )
    end
    add_available_filter(
      "user_id",
      :type => :list_optional, :values => lambda { author_values },
    )
    add_available_filter(
      "counter_party_id",
      :type => :list,
      :values => lambda { entities.map { |t| [t.name, t.id.to_s] } },
    )
    add_available_filter(
      "date_valid_from", :type => :date_past,
    )
    add_available_filter(
      "date_valid_to", :type => :date_past,
    )
    add_available_filter(
      "reward_unit", :type => :list, :values => lambda { reward_unit_values },
    )
    add_available_filter(
      "unit_price", :type => :float,
    )
    add_available_filter(
      "original_currency", :type => :text,
    )
    add_available_filter(
      "original_unit_price", :type => :float,
    )
    add_available_filter(
      "exchange_rate", :type => :float,
    )
    add_available_filter(
      "created_by",
      :type => :list_optional, :values => lambda { author_values },
    )
    add_available_filter(
      "updated_by",
      :type => :list_optional, :values => lambda { author_values },
    )
    add_available_filter "created_at", :type => :date_past
    add_available_filter "updated_at", :type => :date_past
    add_associations_custom_fields_filters :project, :user, :counter_party

  end

  def available_columns
    return @available_columns if @available_columns
    @available_columns = self.class.available_columns.dup
    @available_columns
  end

  def default_columns_names
    @default_columns_names ||= begin
        default_columns = Setting.plugin_redmine_cost_admin["contract_list_defaults"].map(&:to_sym)
      end
  end

  def default_sort_criteria
    [["contract_identifier", "desc"]]
  end

  def base_scope
    Contract.visible.
      joins(:user, :counter_party).
      left_outer_joins(:project).
      where(statement)
  end

  def results_scope(options = {})
    order_option = [group_by_sort_order, (options[:order] || sort_clause)].flatten.reject(&:blank?)

    order_option << "#{Contract.table_name}.id ASC"
    base_scope.
      order(order_option).
      joins(joins_for_order_statement(order_option.join(",")))
  end

  # Returns the contracts count
  def contract_count
    base_scope.count
  rescue ::ActiveRecord::StatementInvalid => e
    raise StatementInvalid.new(e.message)
  end

  # Returns the contracts
  # Valid options are :order, :offset, :limit, :include, :conditions
  def contracts(options = {})
    order_option = [group_by_sort_order, (options[:order] || sort_clause)].flatten.reject(&:blank?)
    # The default order of IssueQuery is issues.id DESC(by IssueQuery#default_sort_criteria)
    unless ["#{Contract.table_name}.id ASC", "#{Contract.table_name}.id DESC"].any? { |i| order_option.include?(i) }
      order_option << "#{Contract.table_name}.id DESC"
    end

    scope = Contract.visible.
      joins(:user, :counter_party).
      left_outer_joins(:project).
      where(statement).
      includes(([:project, :user, :counter_party] + (options[:include] || [])).uniq).
      where(options[:conditions]).
      order(order_option).
      joins(joins_for_order_statement(order_option.join(","))).
      limit(options[:limit]).
      offset(options[:offset])

    scope = scope.preload([:project, :user, :counter_party] & columns.map(&:name))

    contracts = scope.to_a

    contracts
  rescue ::ActiveRecord::StatementInvalid => e
    raise StatementInvalid.new(e.message)
  end



  # Accepts :from/:to params as shortcut filters
  def build_from_params(params, defaults = {})
    super
    if params[:from].present? && params[:to].present?
      add_filter("spent_on", "><", [params[:from], params[:to]])
    elsif params[:from].present?
      add_filter("spent_on", ">=", [params[:from]])
    elsif params[:to].present?
      add_filter("spent_on", "<=", [params[:to]])
    end
    self
  end

  def joins_for_order_statement(order_options)
    joins = [super]

    if order_options
      if order_options.include?("counter_party")
        joins << "LEFT OUTER JOIN #{Entity.table_name} ON #{Entity.table_name}.id = #{Contract.table_name}.counter_party_id"
      end
    end

    joins.compact!
    joins.any? ? joins.join(" ") : nil
  end

  def contract_type_values
    Contract::CONTRACT_TYPE_OPTIONS.map { |n, l| [l(l), n] }
  end

  def reward_unit_values
    Contract::REWARD_UNIT_OPTIONS.map { |n, l| [l(l), n] }
  end

  def entities
    @entities ||= Entity.all
  end
end
