# frozen_string_literal: true

# Redmine - project management software
# Copyright (C) 2006-2019  Jean-Philippe Lang
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

class ExpenseQuery < Query
  self.queried_class = Expense
  self.view_permission = :view_expenses

  self.available_columns = [
    QueryColumn.new(:project, :sortable => "#{Project.table_name}.name", :groupable => true),
    QueryColumn.new(:expense_identifier, :sortable => "#{Project.table_name}.expense_identifier"),
    QueryColumn.new(:date_expensed, :sortable => "#{Expense.table_name}.date_expensed", :groupable => true),
    QueryColumn.new(:amount, :sortable => "#{Expense.table_name}.amount", :totalable => true),
    QueryColumn.new(:description, :sortable => "#{Expense.table_name}.description", :groupable => true),
    TimestampQueryColumn.new(:created_at, :sortable => "#{Expense.table_name}.created_at", :default_order => "desc", :groupable => true),
    TimestampQueryColumn.new(:updated_at, :sortable => "#{Expense.table_name}.updated_at", :default_order => "desc", :groupable => true),
    QueryColumn.new(:created_by, :sortable => "#{Expense.table_name}.created_by", :groupable => true),
    QueryColumn.new(:updated_by, :sortable => "#{Expense.table_name}.updated_by", :groupable => true),
  ]

  def initialize(attributes = nil, *args)
    super attributes
    self.filters ||= { "project" => { :operator => "*", :values => [] } }
  end

  def initialize_available_filters
    add_available_filter(
      "project_id",
      :type => :list, :values => lambda { project_values },
    ) if project.nil?
    if project && !project.leaf?
      add_available_filter(
        "subproject_id",
        :type => :list_subprojects,
        :values => lambda { subproject_values },
      )
    end
    add_available_filter(
      "expense_identifier", :type => :text,
    )
    add_available_filter(
      "date_expensed", :type => :date_past,
    )
    add_available_filter(
      "amount", :type => :float,
    )
    add_available_filter(
      "description", :type => :text,
    )
    add_available_filter(
      "created_by",
      :type => :list_optional, :values => lambda { author_values },
    )
    add_available_filter(
      "updated_by",
      :type => :list_optional, :values => lambda { author_values },
    )
    add_available_filter "created_at", :type => :date_past
    add_available_filter "updated_at", :type => :date_past
    add_associations_custom_fields_filters :project
  end

  def available_columns
    return @available_columns if @available_columns
    @available_columns = self.class.available_columns.dup
    @available_columns
  end

  def default_columns_names
    @default_columns_names ||= begin
        default_columns = Setting.plugin_redmine_cost_admin["expense_list_defaults"].map(&:to_sym)
      end
  end

  def default_totalable_names
    [:amount]
  end

  # Returns sum of all the spent hours
  def total_for_amount(scope)
    map_total(scope.sum(:amount)) { |t| t.to_f.round(2) }
  end

  def default_sort_criteria
    [["date_expensed", "desc"]]
  end

  def base_scope
    Expense.visible.
      joins(:project).
      where(statement)
  end

  def results_scope(options = {})
    order_option = [group_by_sort_order, (options[:order] || sort_clause)].flatten.reject(&:blank?)

    order_option << "#{Expense.table_name}.id ASC"
    base_scope.
      order(order_option).
      joins(joins_for_order_statement(order_option.join(",")))
  end

  # Returns the expenses count
  def expense_count
    base_scope.count
  rescue ::ActiveRecord::StatementInvalid => e
    raise StatementInvalid.new(e.message)
  end

  # Returns the expenses
  # Valid options are :order, :offset, :limit, :include, :conditions
  def expenses(options = {})
    order_option = [group_by_sort_order, (options[:order] || sort_clause)].flatten.reject(&:blank?)
    # The default order of IssueQuery is issues.id DESC(by IssueQuery#default_sort_criteria)
    unless ["#{Expense.table_name}.id ASC", "#{Expense.table_name}.id DESC"].any? { |i| order_option.include?(i) }
      order_option << "#{Expense.table_name}.id DESC"
    end

    scope = Expense.visible.
      joins(:project).
      where(statement).
      includes(([:project] + (options[:include] || [])).uniq).
      where(options[:conditions]).
      order(order_option).
      joins(joins_for_order_statement(order_option.join(","))).
      limit(options[:limit]).
      offset(options[:offset])

    scope = scope.preload([:project] & columns.map(&:name))

    expenses = scope.to_a

    expenses
  rescue ::ActiveRecord::StatementInvalid => e
    raise StatementInvalid.new(e.message)
  end

  # Accepts :from/:to params as shortcut filters
  def build_from_params(params, defaults = {})
    super
    if params[:from].present? && params[:to].present?
      add_filter("date_expensed", "><", [params[:from], params[:to]])
    elsif params[:from].present?
      add_filter("date_expensed", ">=", [params[:from]])
    elsif params[:to].present?
      add_filter("date_expensed", "<=", [params[:to]])
    end
    self
  end

  # def joins_for_order_statement(order_options)
  #   joins = [super]

  #   if order_options
  #     if order_options.include?("counter_party")
  #       joins << "LEFT OUTER JOIN #{Entity.table_name} ON #{Entity.table_name}.id = #{Expense.table_name}.counter_party_id"
  #     end
  #   end

  #   joins.compact!
  #   joins.any? ? joins.join(" ") : nil
  # end
end
